import OpenGL.GL
import numpy
import copy
import state
import texture



# Aux section
# Auxiliary functions


def translate3D(x,y,z):
    return numpy.matrix( [ [  1.0,  0.0,  0.0,    x ],
                           [  0.0,  1.0,  0.0,    y ],
                           [  0.0,  0.0,  1.0,    z ],
                           [  0.0,  0.0,  0.0,  1.0 ] ] );


def rotation3D_x(angle):
    cos = numpy.cos(angle)
    sin = numpy.sin(angle)
    return numpy.matrix( [ [  1.0,  0.0,  0.0,  0.0 ],
                           [  0.0,  cos, -sin,  0.0 ],
                           [  0.0,  sin,  cos,  0.0 ],
                           [  0.0,  0.0,  0.0,  1.0 ] ] );


def rotation3D_y(angle):
    cos = numpy.cos(angle)
    sin = numpy.sin(angle)
    return numpy.matrix( [ [  cos,  0.0, -sin,  0.0 ],
                           [  0.0,  1.0,  0.0,  0.0 ],
                           [  sin,  0.0,  cos,  0.0 ],
                           [  0.0,  0.0,  0.0,  1.0 ] ] );


def rotation3D_z(angle):
    cos = numpy.cos(angle)
    sin = numpy.sin(angle)
    return numpy.matrix( [ [  cos, -sin,  0.0,  0.0 ],
                           [  sin,  cos,  0.0,  0.0 ],
                           [  0.0,  0.0,  1.0,  0.0 ],
                           [  0.0,  0.0,  0.0,  1.0 ] ] );


def rotation3D(anglex, angley, anglez):
    return rotation3D_z(anglez) * rotation3D_y(angley) * rotation3D_x(anglex)


def rotation3D_inv(anglex, angley, anglez):
    return rotation3D_x(-anglex) * rotation3D_y(-angley) * rotation3D_z(-anglez)


def rotation3D_axis(axis, angle):
    # http://www.gamedev.net/reference/articles/article1199.asp
    c = numpy.cos(angle)
    s = numpy.sin(angle)
    t = 1 - c
    x = axis[0]
    y = axis[1]
    z = axis[2]
    return numpy.matrix( [ [  t*x*x + c,   t*x*y + s*z,  t*x*z - s*y,          0.0 ],
                           [  t*y*x - s*z, t*y*y + c,    t*y*z + s*x,          0.0 ],
                           [  t*z*x + s*y, t*z*y - s*x,  t*z*z + c,            0.0 ],
                           [          0.0,         0.0,          0.0,          1.0 ] ] )


def normalize_vector(v):
    return v / numpy.sqrt( numpy.square(v).sum() )


def triangle_normal(mesh, triangle):
    v0, v1, v2 = [ numpy.array(mesh['vertices'][v]['point']) for v in mesh['triangles'][triangle]['vertices'] ]
    return normalize_vector(numpy.cross( v0 - v1, v2 - v1 ))


def triangle_group_normal(mesh, triangleset):
    return normalize_vector( sum(  triangle_normal(mesh, triangle) for triangle in triangleset  ) )


def vertex_group_centroid(mesh, vertices):
    return numpy.average( [ mesh['vertices'][vertex]['point'] for vertex in vertices ] , axis=0)



# Render section
# Functions to render mesh data


def render(mesh):
    if mesh['texture'] != None:
        OpenGL.GL.glEnable(OpenGL.GL.GL_TEXTURE_2D)
        OpenGL.GL.glDisable( OpenGL.GL.GL_COLOR_ARRAY )
        OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_COORD_ARRAY )
        OpenGL.GL.glBindTexture( OpenGL.GL.GL_TEXTURE_2D, mesh['texture'] )
        OpenGL.GL.glTexCoordPointerd( mesh['texcoordarray'] )
    else:
        OpenGL.GL.glDisable(OpenGL.GL.GL_TEXTURE_2D)
        OpenGL.GL.glEnable( OpenGL.GL.GL_COLOR_ARRAY )
        OpenGL.GL.glDisable( OpenGL.GL.GL_TEXTURE_COORD_ARRAY )
        OpenGL.GL.glColorPointerd(    mesh['colourarray'] )
    OpenGL.GL.glVertexPointerd(   mesh['vertexarray'] )
    OpenGL.GL.glNormalPointerd(   mesh['normalarray'] )
    OpenGL.GL.glDrawElementsui(OpenGL.GL.GL_TRIANGLES, mesh['arrayindices'])
    OpenGL.GL.glFinish()



# Compiling section
# Functions to compile mesh data


def mesh2arrays(mesh):
    vertices  = []
    colours   = []
    normals   = []
    texcoords = []
    indices   = []
    
    indexmap = {}

    for triangle in mesh['triangles']:
        for i, vertex in enumerate(triangle['vertices']):
            trianglecolor = 'color' in triangle
            triangletexcoord = (('texcoord' in triangle) and (triangle['texcoord'][i] != None))
            if (vertex not in indexmap) or trianglecolor or triangletexcoord:
                if not(trianglecolor or triangletexcoord): indexmap[vertex] = len(vertices)
                indices.append( len(vertices) )

                v = mesh['vertices'][vertex]
                points = mesh['categories'][ v['category'] ]['points']

                r,g,b,a = v.get('color', (1.0, 1.0, 1.0, 1.0)) if not trianglecolor else triangle['color']
                colours.append([r,g,b,a])

                tx, ty = v.get('texcoord', (1.0, 1.0)) if not triangletexcoord else triangle['texcoord'][i]
                texcoords.append([ tx, ty ])

                point = v['point']
                x, y, z = points[ 0, point ], points[ 1, point ], points[ 2, point ]
                vertices.append([ x, y, z ])

                normal = v['normal']
                nx, ny, nz = points[ 0, normal ] - x, points[ 1, normal ] - y, points[ 2, normal ] - z
                normals.append([ nx, ny, nz ])
            else:
                indices.append( indexmap[vertex] )

    return {'texture'      : mesh['texture'] if 'texture' in mesh else None,
            'vertexarray'  : numpy.ascontiguousarray(vertices),
            'colourarray'  : numpy.ascontiguousarray(colours),
            'normalarray'  : numpy.ascontiguousarray(normals),
            'texcoordarray': numpy.ascontiguousarray(texcoords),
            'arrayindices' : numpy.ascontiguousarray(indices) }


def compile(mesh):
    vertices   = copy.deepcopy( mesh['vertices'] )
    categories = copy.deepcopy( mesh['categories'] )
    joints     = copy.deepcopy( mesh['joints'] )

    triangle_normals = dict((triangle, triangle_normal(mesh, triangle))
                             for triangle, _ in enumerate(mesh['triangles']))

    for category in categories:
        category['points'] = []

    for index, vertex in enumerate(vertices):
        points = categories[ vertex['category'] ]['points']
        x, y, z = vertex['point']
        points.append( [ x, y, z, 1.0 ] )
        vertex['point'] = len( points ) - 1
        if 'normal' in vertex:
            nx, ny, nz = vertex['normal']
        else:
            list = [(triangle_normals[triangle],
                       (mesh['triangles'][triangle]['weight'],
                        mesh['triangles'][triangle]['weight'],
                        mesh['triangles'][triangle]['weight']))
                              for triangle in mesh['vertices'][index]['adjacenttriangles']]
            normal_list, weight_list = zip(*list)
            nx, ny, nz = normalize_vector( numpy.average(normal_list, axis=0, weights=weight_list) )
        points.append( [ nx + x, ny + y, nz + z, 1.0 ] )
        vertex['normal'] = len( points ) - 1

    for category in categories:
        category['points'] = numpy.ascontiguousarray( numpy.matrix( category['points'] ).transpose() )

    for _, joint in joints.iteritems():
        anglex, angley, anglez, length = joint['anglex'], joint['angley'], joint['anglez'], joint['length']
        joint['matrix'] = translate3D( 0.0, 0.0,  length ) * rotation3D_inv(anglex, angley, anglez)

    return state.joindicts( mesh, { 'vertices': vertices, 'categories': categories, 'joints': joints } )


def pose(mesh, pose):
    categories = copy.deepcopy( mesh['categories'] )

    bonematrices = {}
    parentmatrices = {}
    remaining = len( pose )
    while remaining > 0:
        for joint in mesh['joints'].iterkeys():
            if ((joint not in bonematrices)
                 and (('parent' not in mesh['joints'][joint])
                  or (mesh['joints'][joint]['parent'] in parentmatrices))):
                if 'parent 'not in mesh['joints'][joint]:
                    parenttransform = numpy.identity(4, float)
                    parentinverse   = parenttransform
                else:
                    parenttransform, parentinverse = parentmatrices[ mesh['joints'][joint]['parent'] ]
                anglex, angley, anglez = pose[joint]
                transform              = parentinverse * rotation3D(anglex, angley, anglez)
                matrix                 = mesh['joints'][joint]['matrix']
                bonematrices[joint]    = transform * parenttransform
                parentmatrices[joint]  = ( matrix * parenttransform, transform * numpy.linalg.inv(matrix) )
                remaining             -= 1

    for category in categories:
        matrix = numpy.matrix([ [0.0, 0.0, 0.0, 0.0],
                                [0.0, 0.0, 0.0, 0.0],
                                [0.0, 0.0, 0.0, 0.0],
                                [0.0, 0.0, 0.0, 0.0] ])

        for joint, weight in category['joints']:
            matrix += bonematrices[joint] * weight
        category['points'] = matrix * category['points']

    return mesh2arrays( state.joindicts(mesh, { 'categories': categories }) )


def skin(mesh, cutfunction):

    def getadjacentvertices(mesh):
        return dict( (v, frozenset(v2 for t in vertex['adjacenttriangles']
                                            for v2 in mesh['triangles'][t]['vertices']) - frozenset([v]) )
                      for v,vertex in enumerate(mesh['vertices']) )

    def cutsequence(adjacentvertices, cutpoints):
        neighbors = dict( (v, adjacentvertices[v] & cutpoints) for v in cutpoints )
        endpoints = [ v for v,neighborset in neighbors.iteritems() if len(neighborset) == 1 ]
        if len(endpoints) != 2: return []

        sequence = [endpoints[0]]
        set = frozenset([endpoints[0]])
        for i in range(len(cutpoints) - 1):
            nextset = tuple(neighbors[sequence[-1]] - set)
            if len(nextset) != 1: return []
            next = nextset[0]

            sequence.append( next )
            set = set | frozenset([next])
        return sequence

    def left_triangles(mesh, cutpoints, sequence):
        midpoints = frozenset(sequence[1:-1])
        neighbors = frozenset( t for v in midpoints for t in mesh['vertices'][v]['adjacenttriangles'] )

        next = [t for t in neighbors if sequence[0] in mesh['triangles'][t]['vertices']][0]
        left = frozenset([ next ])
        while True:
            adjacents = frozenset( t for v in mesh['triangles'][next]['vertices']
                                         for t in mesh['vertices'][v]['adjacenttriangles']
                                             if (v not in cutpoints) and
                                                (len(frozenset(mesh['triangles'][next]['vertices'])
                                                   & frozenset(mesh['triangles'][t]['vertices'])) == 2) )
            nextset = (adjacents & neighbors) - left
            if len(nextset) != 1: return left
            next = tuple(nextset)[0]
            left = left | nextset

    def getadjacents(mesh, adjacentvertices):

        def vectorlength(vector):
            return numpy.sqrt( numpy.square( vector ).sum() )

        def cosangle(v1, v2):
            return numpy.dot(v1, v2) / ( vectorlength(v1) * vectorlength(v2) )

        def cotangent(v1, v2):
            cosine_alpha = cosangle(v1, v2)
            return cosine_alpha / numpy.sqrt(1 - cosine_alpha*cosine_alpha)

        def weight(mesh, v1, v2):
            triangles = tuple( mesh['vertices'][v1]['adjacenttriangles']
                             & mesh['vertices'][v2]['adjacenttriangles'])
            v3,v4 = (frozenset(mesh['triangles'][triangles[0]]['vertices'])
                      | frozenset(mesh['triangles'][triangles[1]]['vertices'])) - frozenset([v1,v2])
            p1 = numpy.array(mesh['vertices'][v1]['point'])
            p2 = numpy.array(mesh['vertices'][v2]['point'])
            p3 = numpy.array(mesh['vertices'][v3]['point'])
            p4 = numpy.array(mesh['vertices'][v4]['point'])
            return cotangent(p1 - p3, p2 - p3) + cotangent(p1 - p4, p2 - p4)

        return dict( (v, dict( (v2, weight(mesh, v, v2) )
                                 for v2 in adjacentvertices[v]))
                      for v, _ in enumerate(mesh['vertices']) )

    def getlambdas(adjacencymap):
        totals = dict( (v, sum(weight for weight in adjs.itervalues()) )
                        for v,adjs in adjacencymap.iteritems() )
        return dict( (v, dict( (v2, weight / totals[v]) for v2,weight in adjacencymap[v].iteritems() ))
                      for v,adjs in adjacencymap.iteritems() )

    def cutadjacents(adjacentsmap, sequence, leftverts):
        midpoints = frozenset(sequence[1:-1])
        endpoints = frozenset([sequence[0], sequence[-1]])
        map2 = dict((v, dict((v2,d) for v2,d in adjs.iteritems()
                                      if not((v in midpoints) and (v2 in leftverts)) ) )
                     for v,adjs in adjacentsmap.iteritems() if v not in (leftverts | endpoints))
        for midpoint in midpoints:
            adjacents  = [ (v, d) for v,d in adjacentsmap[midpoint].iteritems() if v in (leftverts | endpoints)]
            adjacents2 = [(-v, d) for v,d in adjacentsmap[midpoint].iteritems() if v in midpoints]
            map2[-midpoint] = dict(adjacents + adjacents2)
        for endpoint in endpoints:
            adjacents  = [  (v, d) for v,d in adjacentsmap[endpoint].iteritems() ]
            adjacents2 = [ (-v, d) for v,d in adjacentsmap[endpoint].iteritems() if v in midpoints ]
            map2[endpoint] = dict(adjacents + adjacents2)
        for leftpoint in leftverts:
            adjacents  = [  (v, d) for v,d in adjacentsmap[leftpoint].iteritems() if v not in midpoints ]
            adjacents2 = [ (-v, d) for v,d in adjacentsmap[leftpoint].iteritems() if v in midpoints ]
            map2[leftpoint] = dict(adjacents + adjacents2)
        return map2

    def initialcoords(weightmap, bordersequence):

        def roundpos(position):
            a = position*numpy.pi*2.0
            return numpy.sin(a), numpy.cos(a)

        def squarepos(position):
            if position < 0.25:
                return 4.0*position, 0.0
            elif position < 0.5:
                return 1.0, 4.0*position - 1.0
            elif position < 0.75:
                return 1.0 - (4.0*position - 2.0), 1.0
            else:
                return 0.0, 1.0 - (4.0*position - 3.0)

        weights = [ 1.0 / weightmap[v][bordersequence[(i+1) % len(bordersequence)]]
                    for i,v in enumerate(bordersequence) ]
        totalweights = sum(weights)
        positions = [ sum(weights[0:i+1]) / totalweights for i,_ in enumerate(weights) ]
        coords = [(v, numpy.array(squarepos(pos))) for v,pos in zip(bordersequence, positions)]
        borderset = frozenset(bordersequence)
        coords2 = [(v, numpy.array([0.0, 0.0])) for v in weightmap if v not in borderset]
        return dict( coords + coords2 )

    def barycentric_mapping(lambdas, borderset, coords):
        bordermap = [ v for v in borderset ]
        innermap  = [ v for v in coords if v not in borderset ]

        m = numpy.matrix([ [ (1.0 if i == j else
                              0.0 if innermap[j] not in lambdas[innermap[i]] else
                              -lambdas[innermap[i]][innermap[j]])
                             for j in range(len(innermap)) ]
                                 for i in range(len(innermap)) ])
        k = numpy.matrix([ [ (0.0 if bordermap[j] not in lambdas[innermap[i]] else
                              lambdas[innermap[i]][bordermap[j]])
                             for j in range(len(bordermap)) ]
                                 for i in range(len(innermap)) ])
        c = numpy.matrix([ coords[v] for v in bordermap ])

        x = numpy.linalg.inv(m)*k*c

        innercoords  = [ (innermap[i], numpy.array(coord)[0]) for i,coord in enumerate(x) ]
        bordercoords = [ (v, coords[v]) for v in borderset ]
        return dict(innercoords + bordercoords)

    def applycoords(mesh, coords):
        return state.joindicts(mesh, {'vertices': [ state.joindicts(vertex, { 'texcoord': coords[v] })
                                                    for v, vertex in enumerate(mesh['vertices'])] })

    def left_triangles_coords(mesh, lefttriangles, coords):

        def setcoords(triangle, coords):
            texcoord = tuple((coords[-v] if -v in coords else None) for v in triangle['vertices'])
            return state.joindicts(triangle, { 'texcoord': texcoord })
        
        return state.joindicts(mesh,
                           {'triangles': [(triangle if t not in lefttriangles else setcoords(triangle, coords) )
                                          for t,triangle in enumerate(mesh['triangles'])]})

    cutpoints = frozenset(v for v,vertex in enumerate(mesh['vertices']) if cutfunction(*vertex['point']))
    if len(cutpoints) < 3: return mesh

    adjacentvertices = getadjacentvertices(mesh)
    sequence = cutsequence(adjacentvertices, cutpoints)
    if not sequence: return mesh

    lefttriangles = left_triangles(mesh, cutpoints, sequence)
    leftverts = frozenset(v for t in lefttriangles for v in mesh['triangles'][t]['vertices']) - cutpoints
    adjacents = cutadjacents( getadjacents(mesh, adjacentvertices), sequence, leftverts)

    backsequence = [-v for v in sequence[1:-1]]
    backsequence.reverse()
    roundsequence = sequence + backsequence
    sequenceset = frozenset(roundsequence)
    coords = initialcoords(adjacents, roundsequence)
    lambdas = getlambdas(adjacents)

    coords2 = barycentric_mapping( lambdas, sequenceset, coords )

    coordsmesh = applycoords(mesh, coords2)
    return left_triangles_coords( coordsmesh, lefttriangles, coords2 )


def apply3Dtexture(mesh, texfunc, hash_size = 32, texturesize = 32):

    def get_triangles(mesh):

        def coord(mesh, triangle, index):
            vertex  = mesh['vertices'][ triangle['vertices'][index] ]
            coord2D = numpy.array(triangle['texcoord'][index]
                                   if (('texcoord' in triangle) and triangle['texcoord'][index] != None)
                                   else vertex['texcoord'])
            coord3D = ( numpy.array( vertex['point'] ) if type(vertex['point']) != int
                        else numpy.array(
                                          mesh['categories'][ vertex['category'] ]['points'][:,vertex['point']]
                                        ).transpose()[0] )

            return ( coord2D , coord3D )

        def coords(mesh, triangle):
            return tuple(coord(mesh, triangle, i) for i,_ in enumerate(triangle['vertices']))

        return (coords(mesh, triangle) for triangle in mesh['triangles'])

    def compile_triangles(triangles):

        def compile_triangle(triangle):
            (v1t, v1), (v2t, v2), (v3t, v3) = triangle

            top       = min( v1t[1], v2t[1], v3t[1] )
            left      = min( v1t[0], v2t[0], v3t[0] )
            right     = max( v1t[0], v2t[0], v3t[0] )
            bottom    = max( v1t[1], v2t[1], v3t[1] )
            rectangle = ( numpy.array([left, top]), numpy.array([right, bottom]) )

            t_inv    = numpy.linalg.inv( numpy.matrix([v1t - v3t, v2t - v3t]).transpose() )
            r3       = numpy.matrix(v3t).transpose()
            barycenter = (t_inv, r3)

            matrix3d = numpy.matrix([v1,v2,v3]).transpose()

            return ( rectangle, barycenter, matrix3d )

        return (compile_triangle(triangle) for triangle in triangles)

    def hash_triangles(triangles, nsquares):

        def generate_rows(triangles, nsquares):
            tolerance = 1.0e-13
            rows = [[] for _ in range(nsquares)]
            for triangle in triangles:
                (topleft, rightbottom), _, _ = triangle
                top    = int(numpy.floor(topleft[1]     * nsquares + tolerance))
                bottom = int(numpy.floor(rightbottom[1] * nsquares - tolerance))
                for position in range(top, bottom+1):
                    rows[position].append(triangle)
            return rows

        def generate_columns(row, nsquares):
            tolerance = 1.0e-13
            columns = [[] for _ in range(nsquares)]
            for triangle in row:
                (topleft, rightbottom), _, _ = triangle
                left  = int(numpy.floor(topleft[0]     * nsquares + tolerance))
                right = int(numpy.floor(rightbottom[0] * nsquares - tolerance))
                for position in range(left, right+1):
                    columns[position].append(triangle)
            return columns

        return [ generate_columns(row, nsquares) for row in generate_rows(triangles, nsquares) ]

    def map_2d_to_3d(trianglehash, v):

        def map_point_to_3d(triangle, v):
            (topleft, rightbottom), (t_inv, r3), matrix3d = triangle
            if (topleft > v).any() or (rightbottom < v).any(): return None

            r = numpy.matrix(v).transpose()
            lambdas = t_inv*(r - r3)
            weights = numpy.array([lambdas[0,0], lambdas[1,0], 1.0 - lambdas[0,0] - lambdas[1,0] ])
            tolerance = 1.0e-13
            if (weights < (0.0 - tolerance)).any() or (weights > (1.0 + tolerance)).any(): return None

            point = matrix3d * (numpy.matrix(weights).transpose())

            return numpy.array(point.transpose())[0]

        row    = int(numpy.floor(v[1] * len(trianglehash))) 
        if row >= len(trianglehash): row -= 1
        column = int(numpy.floor(v[0] * len(trianglehash[row])))
        if column >= len(trianglehash[row]): column -= 1
        for triangle in trianglehash[row][column]:
            point = map_point_to_3d(triangle, v)
            if point != None: return point
        return None  # happens if the mesh doesn't fill all the texture

    trianglehash = hash_triangles( compile_triangles( get_triangles(mesh) ), hash_size )

    def texfunc2D(x,y):
        points2D = numpy.array([x.transpose()[0], y.transpose()[0]]).transpose()
        points3D = [ map_2d_to_3d(trianglehash, point) for point in points2D ]
        colors   = [texfunc(point[0], point[1], point[2]) if point != None else (1.0,1.0,1.0,1.0)
                    for point in points3D]
        colorarray = numpy.matrix(colors)
        return colorarray[:,0], colorarray[:,1], colorarray[:,2], colorarray[:,3]

    return state.joindicts(mesh, {'texture': texture.texture(texturesize, texfunc2D) })


def apply2Dtexture(mesh, tex):
    return state.joindicts(mesh, {'texture': tex })


# Construction section
# Functions to construct meshes


def createmesh(categories, joints):
    return { 'vertices': [], 'triangles': [], 'categories': categories, 'joints': joints }


def addvertices(mesh, vertices):
    newvertices = [ state.joindicts(v, {'adjacenttriangles': frozenset([])})
                    for v in vertices ]
    return state.joindicts( mesh, { 'vertices': mesh['vertices'] + newvertices } )


def addtriangles(mesh, triangles):
    newmesh = state.joindicts( mesh, { 'triangles': mesh['triangles'] + triangles } )

    adjacenttriangles = {}
    numoldtriangles = len(mesh['triangles'])
    for i, triangle in enumerate(triangles):
        for v in triangle['vertices']:
            adjacenttriangles[v] = adjacenttriangles.get(v, []) + [numoldtriangles + i]

    newvertices = [ v if i not in adjacenttriangles
                    else state.joindicts(v,
                                {'adjacenttriangles': v['adjacenttriangles'] | frozenset(adjacenttriangles[i]) })
                    for i,v in enumerate(mesh['vertices']) ]

    return state.joindicts( newmesh, {'vertices': newvertices } )


def newmesh(vertices, triangles, categories, joints):
    mesh = createmesh(categories, joints)
    meshv = addvertices(mesh, vertices)
    return addtriangles(meshv, triangles)


def remapvertices(mesh, triangleset, vertexmap):
    newtriangles = [
        triangle if t not in triangleset
                 else state.joindicts( triangle,
                   {'vertices': [ v if v not in vertexmap else vertexmap[v] for v in triangle['vertices'] ] } )
        for t, triangle in enumerate(mesh['triangles']) ]

    newmesh = state.joindicts( mesh, { 'triangles': newtriangles } )

    oldverts = frozenset( vertexmap )
    newverts = frozenset( vertexmap.values() )
    inversemap = dict([ (newvert,oldvert) for oldvert,newvert in vertexmap.iteritems() ])
    newvertices = [ state.joindicts(v, { 'adjacenttriangles': v['adjacenttriangles'] - triangleset })
                   if i in oldverts else 
                   state.joindicts(v, { 'adjacenttriangles': mesh['vertices'][inversemap[i]]['adjacenttriangles']
                                                  & triangleset }) if i in newverts
                   else v
                   for i,v in enumerate(mesh['vertices']) ]

    return state.joindicts( newmesh, {'vertices': newvertices } )


# Generation section
# Functions to generate meshes


def generate2Dmesh(nverts, function, group1, group2, categories, joints):
    t = numpy.linspace(0.0, 1.0, nverts+1)
    x,y = function(t)
    vertices = []
    triangles = []
    for i in range(nverts):
        vertices.append({ 'point': (x[i], y[i], 0.0), 'category': 0 })
    for i in range((nverts - 1) / 2):
        triangles.append( {'vertices': (i+1, i, nverts - i - 1), 'weight': 1.0, 'group': group1 } )
        triangles.append( {'vertices': (nverts - i - 1, i, i+1), 'weight': 1.0, 'group': group2 } )
        if (((nverts - 1) / 2) - i) > (nverts % 2):
            triangles.append( {'vertices': (i+1, nverts-i-1, nverts-i-2), 'weight': 1.0, 'group': group1 } )
            triangles.append( {'vertices': (nverts-i-2, nverts-i-1, i+1), 'weight': 1.0, 'group': group2 } )
    return newmesh( vertices, triangles, categories, joints )


def selectgroup(mesh, group):
    return frozenset([ i for i, t in enumerate(mesh['triangles']) if((group == None) or (t['group'] == group)) ])


def selectedgroup_vertices(mesh, triangleset):
    return frozenset([ v for triangle in triangleset for v in mesh['triangles'][triangle]['vertices'] ])


def select_by_normal(mesh, group, normal, threshold):
    triangleset = selectgroup(mesh, group)
    return frozenset([ t for t in triangleset  if numpy.dot(triangle_normal(mesh, t), normal) > threshold])


def setgroup(mesh, triangleset, newgroup):
    return state.joindicts(mesh, {'triangles': [ state.joindicts(triangle, {'group': newgroup })
                                                 if t in triangleset else triangle
                                                 for t,triangle in enumerate(mesh['triangles']) ]})


def regroup_by_normal(mesh, group, normal, threshold, newgroup):
    triangleset = select_by_normal(mesh, group, normal, threshold)
    return setgroup(mesh, triangleset, newgroup)


def remap_trianglegroups(mesh, groupmap):
    return state.joindicts( mesh, {
                                'triangles': [ t if (('group' not in t) or (t['group'] not in groupmap)) 
                                               else state.joindicts(t, {'group': groupmap[t['group']]})
                                               for t in mesh['triangles'] ]
                            } )


def remap_trianglecolors(mesh, colormap):
    return state.joindicts( mesh, {
                                'triangles': [ t if (('group' not in t) or (t['group'] not in colormap)) 
                                               else state.joindicts(t, {'color': colormap[t['group']]})
                                               for t in mesh['triangles'] ]
                            } )


def vertexgroup_transform_coordinates(mesh, vertexset, function):
    return state.joindicts( mesh, {
                              'vertices': [ v if i not in vertexset
                                            else state.joindicts(v, {'point': function(*v['point']) })
                                            for i,v in enumerate(mesh['vertices']) ]
                            } )


def vertexgroup_transform_displace(mesh, vertexgroup, direction):
    v = numpy.array(direction)

    def displacer(x,y,z):
        return (x,y,z) + v

    return vertexgroup_transform_coordinates(mesh, vertexgroup, displacer)


def vertexgroup_transform_scale(mesh, vertexgroup, factor):
    centroid = vertex_group_centroid(mesh, vertexgroup)

    def scale(x,y,z):
        return ((x,y,z) - centroid)*factor + centroid

    return vertexgroup_transform_coordinates(mesh, vertexgroup, scale)


def vertexgroup_transform_radiate(mesh, vertexgroup, distance):
    centroid = vertex_group_centroid(mesh, vertexgroup)

    def radiate(x,y,z):
        radius = normalize_vector( (x,y,z) - centroid )
        return (x,y,z) + radius*distance

    return vertexgroup_transform_coordinates(mesh, vertexgroup, radiate)


def vertexgroup_transform_rotate(mesh, vertexgroup, axis, angle):
    centroid = vertex_group_centroid(mesh, vertexgroup)
    m = (translate3D( centroid[0], centroid[1], centroid[2] )
        * rotation3D_axis(axis, angle)
        * translate3D( -centroid[0], -centroid[1], -centroid[2] ) )

    def rotate(x,y,z):
        result = m * numpy.array([x,y,z,1.0]).reshape(4,1)
        return result[0,0], result[1,0], result[2,0]

    return vertexgroup_transform_coordinates(mesh, vertexgroup, rotate)


def cutgroup(mesh, triangleset, tubegroup):

    def isborder(mesh, triangleset, vertex1, vertex2):
        adj1 = mesh['vertices'][vertex1]['adjacenttriangles']
        adj2 = mesh['vertices'][vertex2]['adjacenttriangles']
        return len((adj1 & adj2) - triangleset) == 1

    def triangleborder(mesh, triangle, triangleset):
        vertices = mesh['triangles'][triangle]['vertices']
        nverts = len(vertices)
        return [ (vertices[i], vertices[(i+1) % nverts])
                 for i in range(nverts) if isborder(mesh, triangleset, vertices[i], vertices[(i+1) % nverts] ) ]

    def getborder(mesh, triangleset):
        return frozenset([x for triangle in triangleset for x in triangleborder(mesh, triangle, triangleset) ])

    vertexset = selectedgroup_vertices(mesh, triangleset)
    border = getborder(mesh, triangleset)
    bordervertices = frozenset([v for x in border for v in x])    

    # generate the new vertices, as well as the "old vertex position -> new vertex position" map
    oldnvertices = len(mesh['vertices'])
    newvertexdata = [ ((oldpos, newpos + oldnvertices), copy.copy( mesh['vertices'][oldpos] ))
                      for newpos, oldpos in enumerate(bordervertices) ]
    newvertexmap, newvertices = zip(*newvertexdata) # unzip the data
    newvertexmap = dict(newvertexmap)

    newverticesmesh = addvertices(mesh, newvertices)

    bordertriangles = frozenset([ t for v in bordervertices
                                  for t in mesh['vertices'][v]['adjacenttriangles'] ]) - triangleset

    remappedmesh = remapvertices(newverticesmesh, bordertriangles, newvertexmap)

    newtriangles = ( [ {'vertices': (v1, newvertexmap[v2], v2), 'weight': 1.0, 'group': tubegroup }
                     for v1, v2 in border ]
                   + [ {'vertices': (v1, newvertexmap[v1], newvertexmap[v2]), 'weight': 1.0, 'group': tubegroup }
                     for v1,v2 in border] )

    return addtriangles(remappedmesh, newtriangles)


def transform_extrude(mesh, group, distance, tubegroup):
    triangleset = selectgroup(mesh, group)
    normal = triangle_group_normal(mesh, triangleset)
    cutmesh = cutgroup(mesh, triangleset, tubegroup)
    vertexset = selectedgroup_vertices(cutmesh, triangleset)
    return vertexgroup_transform_displace(cutmesh, vertexset, distance*normal)


def transform_coordinates(mesh, group, function):
    triangleset = selectgroup(mesh, group)
    vertexset = selectedgroup_vertices(mesh, triangleset)
    return vertexgroup_transform_coordinates(mesh, vertexset, function)


def transform_displace(mesh, group, direction):
    triangleset = selectgroup(mesh, group)
    vertexset = selectedgroup_vertices(mesh, triangleset)
    return vertexgroup_transform_displace(mesh, vertexset, direction)


def transform_scale(mesh, group, factor):
    triangleset = selectgroup(mesh, group)
    vertexset = selectedgroup_vertices(mesh, triangleset)
    return vertexgroup_transform_scale(mesh, vertexset, factor)


def transform_radiate(mesh, group, distance):
    triangleset = selectgroup(mesh, group)
    vertexset = selectedgroup_vertices(mesh, triangleset)
    return vertexgroup_transform_radiate(mesh, vertexset, distance)


def transform_rotate(mesh, group, axis, angle):
    triangleset = selectgroup(mesh, group)
    vertexset = selectedgroup_vertices(mesh, triangleset)
    return vertexgroup_transform_rotate(mesh, vertexset, axis, angle)


def transform_multiextrude(mesh, group, steps):
    triangleset = selectgroup(mesh, group)
    for step in steps:
        tubegroup, distance, radiate, axis, angle = step
        cutmesh = cutgroup(mesh, triangleset, tubegroup)
        vertexset = selectedgroup_vertices(cutmesh, triangleset)
        normal = triangle_group_normal(cutmesh, triangleset)
        dispmesh = (vertexgroup_transform_displace(cutmesh, vertexset, distance*normal)
                      if distance != 0.0 else cutmesh )
        radmesh  = vertexgroup_transform_radiate(dispmesh, vertexset, radiate) if radiate != 0.0 else dispmesh
        mesh     = vertexgroup_transform_rotate(radmesh, vertexset, axis, angle) if angle != 0.0 else radmesh
    return mesh
