import OpenGL.GL
import numpy
import pygame
import string



# OpenGL section
# Functions to create OpenGL textures from functions


def texture(size, function):
    width   = size
    height  = size
    nelems  = width*height
    x = numpy.linspace(0.0, 1.0, width).reshape(1,width).repeat(height, axis=0).reshape(nelems,1)
    y = numpy.linspace(0.0, 1.0, height).reshape(height,1).repeat(width, axis=1).reshape(nelems,1)

    r, g, b, a = function(x,y)

    v = numpy.concatenate( (r,g,b,a), axis=1 ).reshape(1, 4*nelems)
    pixels = numpy.array( 255*v, dtype='B' ).tostring()

    OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_2D )
    OpenGL.GL.glEnable( OpenGL.GL.GL_BLEND )
    OpenGL.GL.glBlendFunc( OpenGL.GL.GL_SRC_ALPHA, OpenGL.GL.GL_ONE_MINUS_SRC_ALPHA )

    tex = OpenGL.GL.glGenTextures(1)
    OpenGL.GL.glBindTexture( OpenGL.GL.GL_TEXTURE_2D, tex )
    OpenGL.GL.glTexParameteri( OpenGL.GL.GL_TEXTURE_2D,
                               OpenGL.GL.GL_TEXTURE_MAG_FILTER,  OpenGL.GL.GL_LINEAR)
    OpenGL.GL.glTexParameteri( OpenGL.GL.GL_TEXTURE_2D,
                               OpenGL.GL.GL_TEXTURE_MIN_FILTER,  OpenGL.GL.GL_LINEAR)
    OpenGL.GL.glTexImage2D( OpenGL.GL.GL_TEXTURE_2D, 0, OpenGL.GL.GL_RGBA, width, height, 0,
                            OpenGL.GL.GL_RGBA, OpenGL.GL.GL_UNSIGNED_BYTE, pixels)
    return tex


def direct_texture(Surface):
    width, height = Surface.get_size()
    pixels = pygame.image.tostring(Surface, "RGBA", 1)

    OpenGL.GL.glEnable( OpenGL.GL.GL_TEXTURE_2D )
    OpenGL.GL.glEnable( OpenGL.GL.GL_BLEND )
    OpenGL.GL.glBlendFunc( OpenGL.GL.GL_SRC_ALPHA, OpenGL.GL.GL_ONE_MINUS_SRC_ALPHA )

    tex = OpenGL.GL.glGenTextures(1)
    OpenGL.GL.glBindTexture( OpenGL.GL.GL_TEXTURE_2D, tex )
    OpenGL.GL.glTexParameteri( OpenGL.GL.GL_TEXTURE_2D,
                               OpenGL.GL.GL_TEXTURE_MAG_FILTER,  OpenGL.GL.GL_LINEAR)
    OpenGL.GL.glTexParameteri( OpenGL.GL.GL_TEXTURE_2D,
                               OpenGL.GL.GL_TEXTURE_MIN_FILTER,  OpenGL.GL.GL_LINEAR)
    OpenGL.GL.glTexImage2D( OpenGL.GL.GL_TEXTURE_2D, 0, OpenGL.GL.GL_RGBA, width, height, 0,
                            OpenGL.GL.GL_RGBA, OpenGL.GL.GL_UNSIGNED_BYTE, pixels)
    return tex


def closetexture(texture):
    OpenGL.GL.glDeleteTextures(texture)



# Direct section
# Functions that generate textures without combinators


def direct_picture(filename):
    return direct_texture( pygame.image.load(filename) )


def direct_text(s, color, textsize, size):
    color255 = tuple(int(255*col) for col in color)
    font = pygame.font.Font(None, textsize)
    textlines = [ font.render(line, True, color255) for line in string.split(s, "\n") ]
    Surface = pygame.Surface((size,size), flags=pygame.SRCALPHA)
    totalheight = 0
    for line in textlines:
        _, height = line.get_size()
        totalheight += height
        Surface.blit(line, (0,totalheight))
    return direct_texture( Surface )



# Transformation section
# Functions that transform textures


def scale(texture, factor):
    def function(x,y):
      return texture( x / factor, y / factor )

    return function


def translate(texture, deltax, deltay):
    def function(x,y):
      return texture( x  - deltax, y - deltay )

    return function


def rotate(texture, angle):
    def function(x,y):
      cos = numpy.cos(-angle)
      sin = numpy.sin(-angle)
      return texture( x * cos - y * sin, x * sin + y * cos )

    return function


def clamp(texture):
    def function(x,y):
        p = texture(x,y)
        max = p.max()
        min = p.min()
        slope = 1.0 / (max - min)
        constant = -min * slope
        return slope * p + constant

    return function


def fractalize(basetexture, transformations, iterations):
    texture = basetexture
    for i in range(iterations):
        texture2 = texture
        for transformation in transformations:
            texture2 = overlay( transformation(texture), texture2 )
        texture = texture2
    return texture


def sumoctaves(octaves, basetexture):

    def function(x,y):
        sum = numpy.zeros( x.size )
        factor = 1.0
        for i in range(octaves):
            sum += basetexture( x * factor, y * factor ) / (factor * 2.0)
            factor *= 2
        return sum

    return function



# Combinators section
# Combinators that join simple texturing functions into more complex ones


def overlay(texture1, texture2):
    def function(x,y):
        r1, g1, b1, a1 = texture1(x,y)
        r2, g2, b2, a2 = texture2(x,y)
        r = r1 * a1 + (1 - a1) * r2
        g = g1 * a1 + (1 - a1) * g2
        b = b1 * a1 + (1 - a1) * b2
        a = numpy.select([a1 > a2, a1 <= a2], [a1, a2])
        return (r,g,b,a)

    return function


def mix(texture1, texture2, alpha):
    oneminusalpha = 1 - alpha

    def function(x,y):
        r1, g1, b1, a1 = texture1(x,y)
        r2, g2, b2, a2 = texture2(x,y)
        r = alpha * r1 + oneminusalpha * r2
        g = alpha * g1 + oneminusalpha * g2
        b = alpha * b1 + oneminusalpha * b2
        a = alpha * a1 + oneminusalpha * a2
        return (r,g,b,a)

    return function


def mask(texture1, texture2, texturemask):

    def function(x,y):
        r1, g1, b1, a1 = texture1(x,y)
        r2, g2, b2, a2 = texture2(x,y)
        alpha          = texturemask(x,y)
        oneminusalpha  = 1 - alpha
        r = alpha * r1 + oneminusalpha * r2
        g = alpha * g1 + oneminusalpha * g2
        b = alpha * b1 + oneminusalpha * b2
        a = alpha * a1 + oneminusalpha * a2
        return (r,g,b,a)

    return function


def map(texture, colormap):

    def function(x,y):
        return colormap( texture(x,y) )

    return function



# Texture library section
# Specific texturing functions


def surface2texture(surface):
    width, height = surface.get_size()
    maxx, maxy = width - 1, height - 1
    pixstr = pygame.image.tostring(surface, "RGBA", 1)
    pixvec = numpy.fromstring(pixstr, dtype='B') / 255.0
    pixels = pixvec.reshape(height*width, 4)

    def function(x,y):
        ycoords = numpy.array( maxy * y, dtype='I')
        xcoords = numpy.array( maxx * x, dtype='I')
        coords = width*ycoords + xcoords;
        colors = pixels[coords].reshape(height*width, 4)
        return numpy.split(colors, 4, axis=1)

    return function


def picture(filename):
    return surface2texture( pygame.image.load(filename) )


def text(s, color, textsize, size):
    color255 = tuple(int(255*col) for col in color)
    font = pygame.font.Font(None, textsize)
    textlines = [ font.render(line, True, color255) for line in string.split(s, "\n") ]
    Surface = pygame.Surface((size,size))
    totalheight = 0
    for line in textlines:
        _, height = line.get_size()
        totalheight += height
        Surface.blit(line, (0,totalheight))
    return surface2texture( Surface )


def gradient(colors):
    def interpolator(height1, value1, height2, value2):
        def function(height):
            alpha = (height - height1) / (height2 - height1)
            return value1 * (1 - alpha) + value2 * alpha
        return function

    def colormap(heights):
        condlist = []
        funclists = ([],[],[],[]) # One list of functions per component r,g,b,a
        for i in range(len(colors) - 1):
            alpha1, color1 = colors[i]
            alpha2, color2 = colors[i+1]
            condlist.append( (heights >= alpha1) & (heights <= alpha2) )
            for j in range(len(funclists)):
                funclists[j].append( interpolator(alpha1, color1[j], alpha2, color2[j]) )
        color = []
        for i in range(len(funclists)):
            color.append( numpy.piecewise( heights, condlist, funclists[i] ).reshape(heights.shape[0],1) )
        return color

    return colormap


def gradienttable(gradients):
    matrix = numpy.array(gradients)
    m = matrix.shape[1]
    n = matrix.shape[0]

    def function(x,y):
        x = numpy.mod( x, m )
        y = numpy.mod( y, n )
        gradient = matrix[ (y.astype(int), x.astype(int)) ]
        return ( gradient[:,0], gradient[:,1] )

    return function


def perlin(gradients):

    def interpolate(u, v0, v1):
        u2  = u * u
        u3  = u2 * u
        return v0 - (3*u2 - 2*u3)*(v0 - v1)

    def function(x,y):
        x, y = x.reshape(x.size), y.reshape(y.size)

        # Corners of the square containing the point
        x0, y0 = numpy.floor(x), numpy.floor(y)
        x1, y1 = x0,      y0 + 1
        x2, y2 = x0 + 1,  y0
        x3, y3 = x0 + 1,  y0 + 1

        # Gradients at the corners
        g0x, g0y = gradients( x0, y0 )
        g1x, g1y = gradients( x1, y1 )
        g2x, g2y = gradients( x2, y2 )
        g3x, g3y = gradients( x3, y3 )

        # Vectors pointing from corners to point
        v0x, v0y = x - x0, y - y0
        v1x, v1y = x - x1, y - y1
        v2x, v2y = x - x2, y - y2
        v3x, v3y = x - x3, y - y3

        # Dot product from vectors and gradients
        d0 = (v0x*g0x + v0y*g0y + 2.0) / 4.0
        d1 = (v1x*g1x + v1y*g1y + 2.0) / 4.0
        d2 = (v2x*g2x + v2y*g2y + 2.0) / 4.0
        d3 = (v3x*g3x + v3y*g3y + 2.0) / 4.0

        # Bicubic interpolation of the dot products
        d01 = interpolate(v0y, d0, d1)
        d23 = interpolate(v2y, d2, d3)
        return interpolate(v0x, d01, d23)

    return function


def tiledperlin(gradients):
    matrix = numpy.array(gradients)
    m = matrix.shape[1]
    n = matrix.shape[0]
    perlinfunc = perlin(gradienttable(gradients))

    def function(x,y):
        return perlinfunc(x * m, y * n)

    return function
